from skyfield.api import load, Star, Topos
from skyfield.data import hipparcos
import datetime as dt
import tabulate
import prompt_toolkit
from prompt_toolkit.validation import Validator, ValidationError
from prompt_toolkit.completion import WordCompleter

# Load Data and Initialize Named Stars
def load_data_and_stars():
    planets = load('de421.bsp')
    with load.open(hipparcos.URL) as f:
        df = hipparcos.load_dataframe(f)

    named_star_dict = {
                'Acamar': 13847, 'Achernar': 7588, 'Acrux': 60718, 'Adhara': 33579, 'Aldebaran': 21421, 'Algol': 14576,
                'Alioth': 62956,
                'Alkaid': 67301, 'Alnair': 109268, 'Alnilam': 26311, 'Alphard': 46390, 'Alphecca': 76267, 'Alpheratz': 677,
                'Altair': 97649,
                'Ankaa': 2081, 'Antares': 80763, 'Arcturus': 69673, 'Atria': 82273, 'Avior': 41037, 'Becrux': 62434,
                'Bellatrix': 25336,
                'Betelgeuse': 27989, 'Canopus': 30438, 'Capella': 24608, 'Deneb': 102098, 'Denebola': 57632, 'Diphda': 3419,
                'Dubhe': 54061,
                'Elnath': 25428, 'Enif': 107315, 'Eltanin': 87833, 'Fomalhaut': 113368, 'Gacrux': 61084, 'Gienah': 102488,
                'Hadar': 68702, 'Hamal': 9884, 'Kaus Australis': 90185, 'Kochab': 72607, 'Markab': 113963, 'Menkent': 68933,
                'Merak': 53910,
                'Miaplacidus': 45238, 'Mirach': 5447, 'Mirfak': 15863, 'Nunki': 92855, 'Peacock': 100751, 'Polaris': 11767,
                'Pollux': 37826,
                'Procyon': 37279, 'Rasalhague': 86032, 'Regulus': 49669, 'Rigel': 24436, 'Rigel Kent': 71683,
                'RigilKentaurus': 71683,
                'Sabik': 84012, 'Schedar': 3179, 'Shaula': 85927, 'Sirius': 32349, 'Spica': 65474, 'Suhail': 44816,
                'Vega': 91262,
                'Zubenelgenubi': 72622,
            }
    return planets, df, named_star_dict

# Celestial Body Selection Helper
def select_celestial_body(body_name, planets, df, named_star_dict):
    body_mappings = {
        'SunLL': 'Sun', 'SunUL': 'Sun',
        'MoonLL': 'Moon', 'MoonUL': 'Moon',
        'Mars': 'Mars', 'Venus': 'Venus',
        'Jupiter': 'Jupiter Barycenter',  # Barycenter for Jupiter
        'Saturn': 'Saturn Barycenter',    # Barycenter for Saturn
        'Uranus': 'Uranus Barycenter',    # Barycenter for Uranus
        'Mercury': 'Mercury'
    }

    # Check if the body is a planet or its barycenter
    if body_name in body_mappings:
        return planets[body_mappings[body_name]]
    
    # Check if the body is a named star
    star_id = named_star_dict.get(body_name)
    if star_id is not None:
        return Star.from_dataframe(df.loc[star_id])

    # Return None if no valid celestial body is found
    return None


class BodyValidator(Validator):
    def __init__(self, named_star_dict):
        self.valid_bodies = ['SunLL', 'SunUL', 'MoonLL', 'MoonUL', 'Mars', 'Venus', 'Jupiter', 'Saturn', 'Uranus', 'Mercury'] + list(named_star_dict.keys())

    def validate(self, document):
        text = document.text
        if text not in self.valid_bodies:
            raise ValidationError(message='Please enter a valid body', cursor_position=len(text))

class EphemerisCalculator:
    def __init__(self, year, month, day, hour, minute, second, body, latitude, longitude, planets, df, named_star_dict):
        self.body = body
        self.latitude = latitude
        self.longitude = longitude
        self.planets = planets
        self.df = df
        self.named_star_dict = named_star_dict

        ts = load.timescale()
        self.datetime = dt.datetime(year, month, day, hour, minute, second, tzinfo=dt.timezone.utc)
        self.t = ts.utc(self.datetime)

    def compute_gha_dec(self):
        celestial_body = select_celestial_body(self.body, self.planets, self.df, self.named_star_dict)
        if not celestial_body:
            print(f"Unknown celestial body: {self.body}")
            return

        position = self.planets['Earth'] + Topos(latitude_degrees=self.latitude, longitude_degrees=self.longitude)
        astro = position.at(self.t).observe(celestial_body).apparent()

        alt, az, distance = astro.altaz()
        ra, dec, _ = astro.radec(epoch='date')
        gha = (self.t.gast - ra.hours) * 15 % 360
        ghaa = self.t.gast * 15

        print(f"{self.body} at {self.datetime} in position {self.latitude}, {self.longitude}")
        print(tabulate.tabulate([
            ['GHA', f"{gha}°"],
            ['DEC', f"{dec}"],
            ['GHAa', f"{ghaa}°"],
            ['RA', f"{ra}"],
            ['ALT', f"{alt}"],
            ['AZ', f"{az}"],
            ['DIST', f"{distance.au:.2f} AU"],
        ]))

def get_user_input(named_star_dict):
    # Define the prompt toolkit completer and validator
    body_completer = WordCompleter(['SunLL', 'SunUL', 'MoonLL', 'MoonUL', 'Mars', 'Venus', 'Jupiter', 'Saturn', 'Uranus', 'Mercury'] + list(named_star_dict.keys()), ignore_case=True)
    body_validator = BodyValidator(named_star_dict)

    # Get input from user
    body = prompt_toolkit.prompt('Select Body: ', completer=body_completer, validator=body_validator)
    year = int(prompt_toolkit.prompt('Year: '))
    month = int(prompt_toolkit.prompt('Month: '))
    day = int(prompt_toolkit.prompt('Day: '))
    hour = int(prompt_toolkit.prompt('Hour: '))
    minute = int(prompt_toolkit.prompt('Minute: '))
    second = int(prompt_toolkit.prompt('Second: '))
    latitude = float(prompt_toolkit.prompt('Latitude: '))
    longitude = float(prompt_toolkit.prompt('Longitude: '))

    return year, month, day, hour, minute, second, body, latitude, longitude

def main():
    planets, df, named_star_dict = load_data_and_stars()

    # Get user input
    year, month, day, hour, minute, second, body, latitude, longitude = get_user_input(named_star_dict)

    # Create EphemerisCalculator instance and compute
    calculator = EphemerisCalculator(year, month, day, hour, minute, second, body, latitude, longitude, planets, df, named_star_dict)
    calculator.compute_gha_dec()

if __name__ == '__main__':
    main()



from setuptools import setup, find_packages
import toml 

# Get requirements from Pipfile
with open('Pipfile') as f:
    required = toml.load(f)['packages'].keys()
    
setup(
    name='Capella',
    version='2.3.2',
    packages=find_packages(),
    description='A Python package for celestial navigation',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    author='Alex Spradling',
    author_email='alexspradling@gmail.com',
    url='https://github.com/AlexSpradling/Capella',
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
    ],
    install_requires=required
)



class TextExtractor:
    def __init__(self, file_path):
        self.file_path = file_path
        self.text_blocks = self._parse_text_file()

    def _parse_text_file(self):
        with open(self.file_path, 'r') as f:
            lines = f.readlines()

        blocks = {}
        current_block_title = None
        current_block_text = []

        for line in lines:
            stripped_line = line.strip()

            # Check if the line is a title (starts with '# ')
            if stripped_line.startswith('# '):
                # If we're already capturing a block, save it and start a new one
                if current_block_title:
                    key = self._generate_key(current_block_title)
                    blocks[key] = ''.join(current_block_text).strip()
                    current_block_text = []

                current_block_title = stripped_line[2:].strip().lower().replace(' ', '_')
            else:
                current_block_text.append(line)

        # Capture the last block if needed
        if current_block_title and current_block_text:
            key = self._generate_key(current_block_title)
            blocks[key] = ''.join(current_block_text).strip()

        return blocks

    @staticmethod
    def _generate_key(title):
        return title.replace(' ', '_').lower()

    def get_text(self, key):
        return self.text_blocks.get(key, None)



import numpy as np
from skyfield.api import Angle
from utilities.celestial_engine import Sight, SightSession, SightReduction, Utilities
import scipy.stats as stats
from ttkbootstrap.dialogs import Messagebox

# takes info from newpageone treeview, creates SightSession instance and Sight instances and then creates SightReduction instance
# adds Fix information to newpageone fix treeview

class CapellaSightReduction():
    error_flag = False
    def __init__(self, info_fields, treeviews, meter):
        self.sight_treeview = treeviews[0]
        self.fix_treeview = treeviews[1]
        self.info_fields = info_fields
        self.meter = meter
        
        self.delete_fix_treeview()
        self.create_sight_session()
        self.create_sights()
        self.create_sight_reduction()
        self.update_meter()
        self.add_fix_to_treeview()
        self.find_bad_sights()
        self.systematic_error_handling()
   
        self.cnav_data_array_wipe()
    
    def delete_fix_treeview(self):
     # delete all items from fix treeview
        for i in self.fix_treeview.get_children():
            self.fix_treeview.delete(i)

    def create_sight_session(self):
        # get info from newpageone DR info fields
        self.session_instance = SightSession(','.join(str(elem.get()) for elem in self.info_fields))

        return self.session_instance
    
    def create_sights(self,):
        # Iterate through treeview and create Sight instances using list comprehension
        self.sight_instances = [Sight(','.join(self.sight_treeview.item(i, 'values'))) for i in self.sight_treeview.get_children()]

        return self.sight_instances
    
    def create_sight_reduction(self):
        # create SightReduction instance
        self.sight_reduction_instance = SightReduction(True)

        return self.sight_reduction_instance
    

    def add_fix_to_treeview(self):

        # tag configure
        self.fix_treeview.tag_configure('main', font=('Arial Bold', 15))

        # add blank first row to treeview for spacing
        self.fix_treeview.insert('', 'end', text='', iid='blank', values='', tags=('main',))
        
        # add fix to treeview
        for i in self.sight_reduction_instance.gui_position_table:
            self.fix_treeview.insert('', 'end', text='', iid=i, values=i, tags=('main',)) 
    
    def update_meter(self):
        obj_fun_value = self.sight_reduction_instance.res.fun

        def calculate_confidence(obj_fun_value, max_obj_value=0.01):
            if obj_fun_value < 0:
                return 100  # Assuming negative values indicate perfect confidence
            elif obj_fun_value > max_obj_value:
                return 0  # No confidence if the value exceeds the maximum
            else:
                # Linear scaling of confidence
                return 100 * (1 - (obj_fun_value / max_obj_value))
            
        self.meter_value = calculate_confidence(obj_fun_value)
        self.meter.configure(value=int(self.meter_value))


    def find_bad_sights(self):
        '''
        This method finds erroneous sights using z-scores and then asks the user if they want to delete the erroneous sight.
        '''
        # find erroneous sights using z-scores
        self.z_scores = stats.zscore(self.sight_reduction_instance.d_array)

        # iterate through z-scores and highlight erroneous sights
        for idx, d in enumerate(self.sight_reduction_instance.d_array):
            if abs(self.z_scores[idx]) > 2.0:
                # get erroneous sight info
                erroneous_body = Sight.body_array[idx]
                # yyyy-mm-dd hh:mm:ss
                erroneous_sighttime = Sight.sight_times[idx].strftime('%Y-%m-%d %H:%M:%S')
                
                # create tag
                self.sight_treeview.tag_configure('red', foreground='red', font=('Arial Bold', 15))

                # highlight erroneous sight
                self.sight_treeview.item(idx, tags=('red',))

                # ask user if they want to delete erroneous sight
                erroneous_answer = Messagebox.show_question(
                    f'{erroneous_body} observation at {erroneous_sighttime} is likely erroneous.\n\n Correct the observation or remove from Sight List, otherwise consider fix and analysis unreliable. \
                    \n\n Would you like to delete this observation?')
                
                # if yes delete erroneous sight
                if erroneous_answer == 'Yes':
                    error_flag = True
                    # set selection 
                    self.sight_treeview.selection_set(idx)

                    # delete erroneous sight
                    for i in self.sight_treeview.selection():
                        self.sight_treeview.delete(i)
                    error_flag = False

    def systematic_error_handling(self):
        """
        This method finds the systematic error and asks the user if they want to remove it.
        """
       
        self.count = 0
        # self.systematic_error = np.mean(self.sight_reduction_instance.d_array)

        # make systematic_error the trimmed mean of d_array
        self.systematic_error = stats.trim_mean(self.sight_reduction_instance.d_array, .1)
        
        message = f"Capella found a Constant Error (Uncorrected Index Error + Personal Error) of: {np.round(self.systematic_error, 2)}'.\
        \n\n\n Would you like to remove this error? "


        if abs(self.systematic_error) >=.25:
            error_flag = True
            error_message = Messagebox.show_question(message, 'Systematic Error',)
            if self.systematic_error < 0:
                self.systematic_error = abs(self.systematic_error / 60)
            else:
                self.systematic_error = self.systematic_error / 60 * -1
            
            if error_message == 'Yes' and self.systematic_error !=0:
                for i in self.sight_treeview.get_children():
                    body = self.sight_treeview.item(i, 'values')[0]
                    hs = self.sight_treeview.item(i, 'values')[1]
                    date = self.sight_treeview.item(i, 'values')[2]
                    time = self.sight_treeview.item(i, 'values')[3]
                    hs_deg, hs_min = hs.split('-')
                    hs = float(hs_deg) + float(hs_min) / 60
                    hs = Angle(degrees=hs)
                    hs = Utilities.hmt_str_2(hs.degrees + self.systematic_error)
                    self.sight_treeview.delete(i)
                    self.sight_treeview.insert('', 'end', text='', iid=i, values=(body, hs, date, time), tags=('main',))

                # wipe data arrays and run CapellaSightReduction again
                self.cnav_data_array_wipe()

                # recursively run CapellaSightReduction
                CapellaSightReduction(self.info_fields, [self.sight_treeview, self.fix_treeview], self.meter)
                self.count += 1
                error_flag = False

        else:
            return

    
    def cnav_data_array_wipe(self):
            
            # Sight_session
            SightSession.num_of_sights = 0
            SightSession.dr_details = []

            # Sight
            Sight.data_table = []
            Sight.sight_times = []
            Sight.num_of_sights = 0
            Sight.body_array = []
            Sight.sight_az_array = []
            Sight.sight_times = []
            Sight.computedlong = []
            Sight.computedlat = []
            Sight.intercept_array = []
            Sight.ho_array = []
            Sight.dec_array_lop = []
            Sight.gha_array_lop = []
            Sight.hc_array = []
            Sight.gha_dec_array = []
            Sight.test_array_ho = []
            Sight.ho_vec_array = []
            Sight.test_array_gha = []

            # Sight_Reduction

            SightReduction.gui_position_table = []
            SightReduction.test_array
            SightReduction.final_position_array = []
            SightReduction.stats_table_2 = []
            SightReduction.latx_lists = []
            SightReduction.longx_lists = []

            SightReduction.ho_array_rfix = []
            SightReduction.time_delta_array = []
            SightReduction.sight_anl_table = []
            SightReduction.final_ho_array = []
            SightReduction.pos_array_lop_lon = []
            SightReduction.pos_array_lop_lat = []
            SightReduction.d_array = []
            SightReduction.ho_corrections_array = []
            SightReduction.longitude_array = []
            SightReduction.latitude_array = []
            SightReduction.hc_timeofsight = []
            SightReduction.sight_analysis_lat_time_of_sight = []
            SightReduction.sight_analysis_long_time_of_sight = []
            SightReduction.sight_analysis_lat_minus_one = []
            SightReduction.sight_analysis_long_minus_one = []
            SightReduction.sight_analysis_lat_plus_one = []
            SightReduction.sight_analysis_long_plus_one = []
            SightReduction.hc_minusone = []
            SightReduction.hc_plusone = []
            SightReduction.position_array_l = []

    